package be.iminds.ilabt.jfed.rspec.rspec_source;

import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableModelRspec;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableRspecFactory;
import be.iminds.ilabt.jfed.rspec.parser.RspecCompare;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecParser;
import be.iminds.ilabt.jfed.rspec.util.ProgressHandler;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.ThreadFactoryUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.stream.XMLStreamException;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 * RspecSource
 * <p>
 * Note: this is a somewhat immutable class
 * <p>
 * Note: this caches a bit too much, and doesn't always copy enough.
 * There is risk of the mutable RSpecModel coming out of this being changed by another.
 * Try using ImmutableRSpecSource instead!
 * <p>
 * RspecSource takes care of converting and caching different StringRspecs and ModelRspecs.
 * You can create one by providing either a StringRspec (assumed ModelRspec.RequestRspecSpecialCases.NONE) or a ModelRspec.
 * Then you can request any type of special case StringRspec, or a ModelRspec for any type of factory.
 * The RspecSource will convert what it has to create the requested rspec if needed, and cache the result so these conversions need to be done as little as possible.
 * <p>
 * This class tries to minimize blocking on RSpec-parsing by performing the parsing in a separate thread.
 * If parsing hasn't completed yet, and a ModelRspec is requested, the call will block on this request until the parsing has completed.
 */
public abstract class CachingRspecSource implements RspecSource {
    private static final Logger LOG = LoggerFactory.getLogger(CachingRspecSource.class);

    private static final ExecutorService executorService = Executors.newCachedThreadPool(
            ThreadFactoryUtil.getFactory("RspecSource"));

    protected final Map<ModelRspec.RequestRspecSpecialCases, StringRspec> stringRspecBySpecialCase =
            new EnumMap<>(ModelRspec.RequestRspecSpecialCases.class);
    protected final Map<ModelRspecType, StaxRspecParser> modelRspecParserMap =
            new EnumMap<>(ModelRspecType.class);
    protected final Map<ModelRspecType, Future<ModelRspec>> modelRspecFutureMap =
            new EnumMap<>(ModelRspecType.class);
    protected final Map<ModelRspecType, ModelRspec> modelRspecMap =
            new EnumMap<>(ModelRspecType.class);

    private boolean initialized = false;
    protected final boolean isXmlBased;

    @Nonnull private final ModelRspecType defaultMutableModelRspecType;

    /* COPY constructor */
    protected CachingRspecSource(@Nonnull RspecSource orig) {
        if (orig.isXmlBased()) {
            this.stringRspecBySpecialCase.put(ModelRspec.RequestRspecSpecialCases.NONE, orig.getStringRspec());
            this.isXmlBased = true;
            this.defaultMutableModelRspecType = ModelRspecType.BASIC;
            this.initialized = false;
        } else {
            ModelRspec origModelRspec = orig instanceof CachingRspecSource ?
                    ((CachingRspecSource) orig).getAnyModelRspec() :
                    orig.getImmutableModelRspec();
            assert origModelRspec != null;
            this.modelRspecMap.put(origModelRspec.getModelRspecType(), origModelRspec);
            this.defaultMutableModelRspecType = ModelRspecType.BASIC;
            this.isXmlBased = false;
            this.initialized = true; // no lazy initialization needed anymore
        }
    }

    protected CachingRspecSource(@Nonnull StringRspec stringRspec) {
        this(stringRspec, ModelRspecType.BASIC);
    }

    /**
     * @param defaultMutableModelRspecType default ModelRspecType to use. may be null. (= BASIC in that case)
     */
    protected CachingRspecSource(@Nonnull StringRspec stringRspec, @Nonnull ModelRspecType defaultMutableModelRspecType) {
        assert stringRspec != null;
        this.stringRspecBySpecialCase.put(ModelRspec.RequestRspecSpecialCases.NONE, stringRspec);
        this.defaultMutableModelRspecType = defaultMutableModelRspecType;
        this.isXmlBased = true;
        this.initialized = false;
    }

    protected CachingRspecSource(@Nonnull ModelRspec modelRspec) {
        assert modelRspec != null;
        this.modelRspecMap.put(modelRspec.getModelRspecType(), modelRspec); //TODO make a copy of modelRspec
        this.defaultMutableModelRspecType = modelRspec.getModelRspecType();
        this.isXmlBased = false;
        this.initialized = true; // no lazy initialization needed anymore
    }

    /**
     * @param defaultMutableModelRspecType default ModelRspecType to use. may be null. (= BASIC in that case)
     */
    public CachingRspecSource(@Nonnull String xmlRspecString, @Nonnull ModelRspecType defaultMutableModelRspecType) {
        this(new StringRspec(xmlRspecString), defaultMutableModelRspecType);
    }

    /**
     * This method is used for lazy initialization of the first ModelRspec
     */
    private void initialize() {
        getAnyModelRspec(); //the RspecSource will be fully initialized after the first ModelRspec has been created
    }

    public boolean isXmlBased() {
        if (!initialized)
            initialize();

        return isXmlBased;
    }

    @Nullable
    public String getRspecXmlString() {
        return getRspecXmlString(ModelRspec.RequestRspecSpecialCases.NONE);
    }

    @Nullable
    public String getRspecXmlString(@Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase) {
        final StringRspec stringRspec = getStringRspec(rspecSpecialCase);
        if (stringRspec != null)
            return stringRspec.getXmlString();
        else return null;
    }

    protected boolean parseFailed = false;

    @Nonnull
    protected String modelToString(@Nonnull ModelRspec modelRspec, @Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase) {
        assert (modelRspec != null);
        return modelRspec.toGeni3Rspec(rspecSpecialCase);
    }


    /**
     * @return a modelRspec, created by any mutable RspecFactory
     */
    @Nullable
    public ModelRspec getMutableModelRspec() {
        ModelRspec modelRspec = getModelRspec(defaultMutableModelRspecType);
        Iterator<Map.Entry<ModelRspecType, ModelRspec>> it = modelRspecMap.entrySet().iterator();
        if ((modelRspec == null || modelRspec.getModelRspecType() == ModelRspecType.IMMUTABLE) && it.hasNext())
            modelRspec = it.next().getValue();
        return modelRspec;
    }

    /**
     * for internal use, when we don't care if it's immutable or not, we just want any model
     * @return
     */
    @Nullable
    protected ModelRspec getAnyModelRspec() {
        ModelRspec modelRspec = getModelRspec(defaultMutableModelRspecType);
        if (modelRspec == null && !modelRspecMap.isEmpty())
            modelRspec = modelRspecMap.entrySet().iterator().next().getValue();
        return modelRspec;
    }

    /**
     * @return a modelRspec, created by any immutable RspecFactory
     */
    @Nullable
    public ImmutableModelRspec getImmutableModelRspec() {
        ModelRspec modelRspec = modelRspecMap.get(ModelRspecType.IMMUTABLE);
        if (modelRspec == null)
            modelRspec = getModelRspec(ModelRspecType.IMMUTABLE);
        return (ImmutableModelRspec) modelRspec;
    }

    /**
     * @return a modelRspec, created by any mutable RspecFactory
     */
    @Nullable
    public ModelRspec getMutableModelRspec(ProgressHandler... progressHandlers) {
        ModelRspec modelRspec = getModelRspec(defaultMutableModelRspecType, progressHandlers);
        Iterator<Map.Entry<ModelRspecType, ModelRspec>> it = modelRspecMap.entrySet().iterator();
        if ((modelRspec == null || modelRspec.getModelRspecType() == ModelRspecType.IMMUTABLE) && it.hasNext())
            modelRspec = it.next().getValue();
        return modelRspec;
    }

    /**
     * @return a modelRspec, created by any Immutable RspecFactory
     */
    @Nullable
    public ImmutableModelRspec getImmutableModelRspec(ProgressHandler... progressHandlers) {
        ModelRspec modelRspec = modelRspecMap.get(ModelRspecType.IMMUTABLE);
        if (modelRspec == null)
            modelRspec = getModelRspec(ModelRspecType.IMMUTABLE, progressHandlers);
        return (ImmutableModelRspec) modelRspec;
    }

    public boolean isModelRspecAvailable(@Nonnull ModelRspecType modelRspecType) {
        return modelRspecMap.containsKey(modelRspecType);
    }


    @SuppressWarnings("unchecked")
    @Nullable
    public <M extends ModelRspec> M getModelRspec(@Nonnull ModelRspecType modelRspecType, ProgressHandler... progressHandlers) {
        if (modelRspecMap.containsKey(modelRspecType))
            return (M)modelRspecMap.get(modelRspecType);
        if (parseFailed)
            return null;

        Future<ModelRspec> future = getModelRspecAsync(modelRspecType == ModelRspecType.IMMUTABLE ? defaultMutableModelRspecType : modelRspecType, progressHandlers);

        ImmutableModelRspec immRes = null; //messy
        try {
            if (future == null) {
                throw new ExecutionException(
                        new RspecParseException(null, "Cannot create a modelRspec from an empty Rspec"));
            }

            if (modelRspecType == ModelRspecType.IMMUTABLE) {
                ModelRspec res = future.get();
                immRes = ImmutableRspecFactory.getInstance().copyModelRspec(res, null);
                return (M) immRes;
            } else {
                return (M)future.get();
            }
        } catch (ExecutionException e) {
            if (e.getCause() instanceof RspecParseException) {
                lastInvalidRspecException = (RspecParseException) e.getCause();
            }

            LOG.warn("RSpec String parse failed. Will not return ModelRspec", e);
            parseFailed = true;
            return null;
        } catch (InterruptedException e) {
            LOG.error("Rspec parsing failed unexpectedly!", e);
            parseFailed = true;
            return null;
        } finally {
            initialized = true;
            if (future != null && future.isDone()) {
                try {
                    ModelRspec res = future.get();
                    modelRspecMap.put(res.getModelRspecType(), res);
                    if (immRes != null) {
                        modelRspecMap.put(immRes.getModelRspecType(), immRes);
                    }
                } catch (InterruptedException | ExecutionException ignored) {
                }
            }
        }
    }

    @Nullable
    private Future<ModelRspec> getModelRspecAsync(@Nonnull ModelRspecType modelRspecType, ProgressHandler... progressHandlers) {
        Future<ModelRspec> future;
        synchronized (this) {
            StaxRspecParser parser;
            if (!modelRspecParserMap.containsKey(modelRspecType)) {
                //get the correct StringRspec
                StringRspec stringRspec;
                if (stringRspecBySpecialCase.containsKey(ModelRspec.RequestRspecSpecialCases.NONE))
                    stringRspec = stringRspecBySpecialCase.get(ModelRspec.RequestRspecSpecialCases.NONE);
                else {
                    if (!stringRspecBySpecialCase.isEmpty())
                        stringRspec = stringRspecBySpecialCase.get(stringRspecBySpecialCase.keySet().iterator().next());
                    else {
                        //this onlt occurs when we do not have a stingrspec, but we have a modelrspec of the wrong type
                        stringRspec =
                                getStringRspec(); //no stringRspec stored. getStringRspec() will generate one from the modelRspec
                        //this can be done more efficiently, as ModelRspec can be copied. But this is probably the safest method
                    }
                }

                if (stringRspec == null)
                    return null;

                String xmlString = stringRspec.getXmlString();

                //quick HACK to allow the planetlab Europe rspec manifest
                if (xmlString.contains("<RSpec")) {
                    LOG.debug("HACK: rewriting received planetlab manifest XML to make it parsable xml");
                    xmlString =
                            xmlString.replace("<RSpec type=\"SFA\"", "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" type=\"manifest\"");
                    xmlString =
                            xmlString.replace("<RSpec ", "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" type=\"manifest\"");
                    xmlString = xmlString.replace("</RSpec>", "</rspec>");
                    xmlString = xmlString.replace("<network name=\"ple\">", "");
                    xmlString = xmlString.replace("</network>", "");
                    LOG.debug("HACK: rewritten xml: " + xmlString);
                }

                parser = new StaxRspecParser(xmlString, RspecFactoryFactory.getRspecFactoryInstance(modelRspecType));
                future = executorService.submit(parser);


            } else {
                parser = modelRspecParserMap.get(modelRspecType);
                future = modelRspecFutureMap.get(modelRspecType);
            }

            //add progress-handlers
            for (ProgressHandler handler : progressHandlers)
                parser.addProgressHandler(handler);
        }
        return future;
    }


    private RspecParseException lastInvalidRspecException = null;

    @Nullable
    public RspecParseException getLastInvalidRspecException() {
        return lastInvalidRspecException;
    }

    @Nullable
    public StringRspec getStringRspec() {
        return getStringRspec(ModelRspec.RequestRspecSpecialCases.NONE);
    }

    @Nullable
    public StringRspec getStringRspec(@Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase) {
        if (stringRspecBySpecialCase.containsKey(rspecSpecialCase))
            return stringRspecBySpecialCase.get(rspecSpecialCase);
        ModelRspec modelRspec = getAnyModelRspec();
        if (modelRspec == null) {
            assert !stringRspecBySpecialCase.isEmpty() :
                    "both modelRspec == null and stringRspecBySpecialCase.isEmpty()";

            try {
                modelRspec = getModelRspec(defaultMutableModelRspecType);
            } catch (AssertionError | Exception t) {
                LOG.warn("RSpec String parse failed. Cannot change type of rspec to " + rspecSpecialCase, t);
                parseFailed = true;
                return null;
            }

            if (modelRspec == null)
                return null;
        }
        assert modelRspec != null;

        String rspecXml = modelToString(modelRspec, rspecSpecialCase);
        StringRspec stringRspec = new StringRspec(rspecXml);
        stringRspecBySpecialCase.put(rspecSpecialCase, stringRspec);
        return stringRspec;
    }


    @Nonnull
    public List<GeniUrn> getAllComponentManagerUrns() {
        ModelRspec model = getAnyModelRspec();
        if (model != null) {
            List<GeniUrn> res = model.getAllComponentManagerUrns();

            StringRspec stringRspec = getStringRspec();

            if (stringRspec == null)
                return Collections.emptyList();

            assert stringRspec.isWellFormed();

            //parse generated rspec for openflow component URNs.
            //   Note: This could be done more efficiently, by parsing "others", but it's not worth the trouble

            for (String urnString : StringRspec.parseOpenflowComponentManagers(stringRspec.getXmlString())) {
                GeniUrn geniUrn = GeniUrn.parse(urnString);

                if (geniUrn != null)
                    res.add(geniUrn);
                else {
                    LOG.warn("Could not parse openflowComponentManager-urn {}", urnString);
                }
            }

            return res;
        } else {
            StringRspec stringRspec = getStringRspec();
            if (stringRspec != null && stringRspec.isWellFormed()) {
                List<GeniUrn> res = new ArrayList<>();
                for (String urnString : stringRspec.getAllComponentManagerUrns()) {
                    GeniUrn geniUrn = GeniUrn.parse(urnString);

                    if (geniUrn != null)
                        res.add(geniUrn);
                    else {
                        LOG.warn("Could not parse openflowComponentManager-urn {}", urnString);
                    }
                }

                return res;
            } else
                return Collections.emptyList();
        }
    }

    public List<BasicStringRspec.BasicNodeInfo> getBasicNodeInfo() {
        ModelRspec model = getAnyModelRspec();

        if (model != null) {
            List<BasicStringRspec.BasicNodeInfo> res = new ArrayList<>();
            List<? extends RspecNode> modelNodes = model.getNodes();

            for (RspecNode node : modelNodes) {
                List<String> sliverTypeNames = new ArrayList<>();
                for (SliverType sliverType : node.getSliverTypes())
                    sliverTypeNames.add(sliverType.getName());
                BasicStringRspec.BasicNodeInfo ni = new BasicStringRspec.BasicNodeInfo(
                        node.getClientId(),
                        node.getComponentId() == null ? null : node.getComponentId().toString(),
                        node.getComponentManagerId() == null ? null : node.getComponentManagerId().toString(),
                        node.getAvailable(),
                        node.getExclusive(),
                        sliverTypeNames,
                        node.getAnsibleGroups());
                res.add(ni);
            }

            return res;
        } else {
            StringRspec stringRspec = getStringRspec();
            LOG.debug("The RSpec could not be parsed using ModelRspec, will try StringRspec method.");
            if (stringRspec != null && stringRspec.isWellFormed()) {
                return stringRspec.getBasicNodeInfo();
            } else {
                LOG.warn("The RSpec could not be parsed using any method. Will return null.");
                return null;
            }
        }
    }

    private final Map<ModelRspecType, LosingDataInfo> losingDataInfoMap = new HashMap<>();

    @Nullable
    private LosingDataInfo getLosingDataInfo(ModelRspecType modelRspecType) {
        if (!isXmlBased)
            return null; //cannot lose data when not xml-based!

        if (!losingDataInfoMap.containsKey(modelRspecType)) {
            ModelRspec modelRspec = getModelRspec(modelRspecType);
            String stringRspec = getRspecXmlString();

            if (modelRspec == null || stringRspec == null)
                return null;

            LosingDataInfo losingDataInfo;
            //set isLosingData, but handle any failure.
            try {
                RspecCompare compareWithOrig =
                        new RspecCompare(modelRspec.toGeni3Rspec(), stringRspec, "the converted RSpec", "the original RSpec");
                losingDataInfo =
                        new LosingDataInfo(!compareWithOrig.isEqual(), compareWithOrig.getDifferences()); //TODO: only calculate and set losingData this when it is actually needed
            } catch (XMLStreamException ex) {
                LOG.error("Exception while using RspecCompare at end of parse. location=" + RspecParseException.locationToString(ex.getLocation()), ex);
                losingDataInfo = new LosingDataInfo(true, /*"Exception while comparing: "+ex.getMessage()*/null);
            } catch (AssertionError | Exception ex) {
                LOG.error("Exception while using RspecCompare at end of parse. location=unknown", ex);
                losingDataInfo = new LosingDataInfo(true, /*"Exception while comparing: "+ex.getMessage()*/null);
            }

            losingDataInfoMap.put(modelRspecType, losingDataInfo);
        }

        return losingDataInfoMap.get(modelRspecType);
    }

    public boolean isLosingData(ModelRspecType modelRspecType) {
        LosingDataInfo losingDataInfo = getLosingDataInfo(modelRspecType);
        return losingDataInfo != null && losingDataInfo.isLosingData();
    }

    public List<RspecCompare.Difference> getLosingDataDifferences(ModelRspecType modelRspecType) {
        LosingDataInfo losingDataInfo = getLosingDataInfo(modelRspecType);
        if (losingDataInfo != null)
            return losingDataInfo.getLosingDataDifferences();
        else
            return null;
    }
}
